from __future__ import print_function
import argparse
from hyperopt import hp
from hyperopt.pyll.stochastic import sample
import pickle
import sys
import os
import time

# Custom generator for our dataset
from modules.python.models.hyperband import Hyperband
from modules.python.TextColor import TextColor
from modules.python.models.train import train
from modules.python.Options import TrainOptions
"""
Tune hyper-parameters of a model using hyperband.
Input:
- A train CSV file
- A test CSV file

Output:
- A model with tuned hyper-parameters
"""


class WrapHyperband:
    """
    Wrap hyperband around a model to tune hyper-parameters.
    """
    # Paramters of the model
    # depth=28 widen_factor=4 drop_rate=0.0
    def __init__(self, train_file, test_file, gpu_mode, model_out_dir, log_dir, max_epochs, batch_size, num_workers):
        """
        Initialize the object
        :param train_file: A train CSV file containing train data set information
        :param test_file: A test CSV file containing train data set information
        :param gpu_mode: If true, GPU will be used to train and test the models
        :param model_out_dir: Directory to save the model
        :param log_dir: Directory to save the log
        """
        # the hyper-parameter space is defined here
        self.space = {
            # hp.loguniform returns a value drawn according to exp(uniform(low, high)) so that the logarithm of the
            # return value is uniformly distributed.
            'lr': hp.loguniform('lr', -8, -4),
            'l2': hp.loguniform('l2', -12, -4)
        }
        self.train_file = train_file
        self.test_file = test_file
        self.gpu_mode = gpu_mode
        self.log_directory = log_dir
        self.model_out_dir = model_out_dir
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.hidden_size = TrainOptions.HIDDEN_SIZE
        self.gru_layers = TrainOptions.GRU_LAYERS

    def get_params(self):
        """
        Get a random draw from the parameter space.
        :return: A randomly drawn sample space
        """
        return sample(self.space)

    def try_params(self, n_iterations, model_params, model_path):
        """
        Try a parameter space to train a model with n_iterations (epochs).
        :param n_iterations: Number of epochs to train on
        :param model_params: Parameter space
        :return: trained model, optimizer and stats dictionary (loss and others)
        """
        # Number of iterations or epoch for the model to train on
        n_iterations = int(round(n_iterations))
        params, retrain_model, retrain_model_path, prev_ite = model_params
        sys.stderr.write(TextColor.BLUE + '\nEpochs: ' + str(n_iterations) + "\n" + TextColor.END)
        sys.stderr.write(TextColor.BLUE + str(params) + "\n" + TextColor.END)

        epoch_limit = int(n_iterations)
        lr = params['lr']
        l2 = params['l2']

        # train a model
        transducer_model, model_optimizer, stats_dictionary = train(self.train_file,
                                                                    self.test_file,
                                                                    self.batch_size,
                                                                    epoch_limit,
                                                                    self.gpu_mode,
                                                                    self.num_workers,
                                                                    retrain_model,
                                                                    retrain_model_path,
                                                                    self.gru_layers,
                                                                    self.hidden_size,
                                                                    lr,
                                                                    l2,
                                                                    model_dir=None,
                                                                    stats_dir=None,
                                                                    train_mode=False)

        # save_best_model(transducer_model, model_optimizer, self.hidden_size, self.gru_layers, epoch_limit, model_path)

        return transducer_model, model_optimizer, stats_dictionary

    def run(self, save_output):
        """
        Run the hyper-parameter tuning algorithm
        :param save_output: If true, output will beb saved in a pkl file
        :return:
        """
        hyperband = Hyperband(self.get_params, self.try_params, max_iteration=self.max_epochs, downsample_rate=3,
                              model_directory=self.model_out_dir, log_directory=self.log_directory)
        results = hyperband.run()

        if save_output:
            with open(self.log_directory + 'results.pkl', 'wb') as f:
                pickle.dump(results, f)

        # Print top 5 configs based on loss
        results = sorted(results, key=lambda r: r['loss'])[:5]
        for i, result in enumerate(results):
            print(i+1)
            print(result)
            print("Loss:\t\t", result['loss'])
            print("Accuracy:\t", result['accuracy'])
            print("Params:\t\t", result['params'])
            print("Model path:\t", result['model_path'])


import sys
import torch

# Custom generator for our dataset
from modules.python.models.train import train
from modules.python.models.train_distributed import train_distributed
from modules.python.Options import TrainOptions
from modules.python.FileManager import FileManager
from modules.python.TextColor import TextColor

"""
The train module of HELEN trains a deep neural network to perform a multi-task classification. It takes a set of
labeled images from MarginPolish and trains the model to predict a base and the run-length of the base using a 
gated recurrent unit (GRU) based model. This script is the interface to the training module. 
"""


class TrainModule:
    """
    Train module that provides an interface to the train method of HELEN.
    """
    def __init__(self, train_file, test_file, gpu_mode, device_ids, max_epochs, batch_size, num_workers,
                 retrain_model, retrain_model_path, model_dir, stats_dir):
        self.train_file = train_file
        self.test_file = test_file
        self.gpu_mode = gpu_mode
        self.device_ids = device_ids
        self.model_dir = model_dir
        self.epochs = max_epochs
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.retrain_model = retrain_model
        self.retrain_model_path = retrain_model_path
        self.stats_dir = stats_dir
        self.hidden_size = TrainOptions.HIDDEN_SIZE
        self.gru_layers = TrainOptions.GRU_LAYERS
        self.learning_rate = 0.0001
        self.weight_decay = 0

    def train_model(self):
        # train a model
        train(self.train_file,
              self.test_file,
              self.batch_size,
              self.epochs,
              self.gpu_mode,
              self.num_workers,
              self.retrain_model,
              self.retrain_model_path,
              self.gru_layers,
              self.hidden_size,
              self.learning_rate,
              self.weight_decay,
              self.model_dir,
              self.stats_dir,
              not_hyperband=True)

    def train_model_gpu(self):
        """
        DO DISTRIBUTED GPU INFERENCE. THIS MODE WILL ENABLE ONE MODEL PER GPU
        """
        if not torch.cuda.is_available():
            sys.stderr.write(TextColor.RED + "ERROR: TORCH IS NOT BUILT WITH CUDA.\n" + TextColor.END)
            sys.stderr.write(TextColor.RED + "SEE TORCH CAPABILITY:\n$ python3\n"
                                             ">>> import torch \n"
                                             ">>> torch.cuda.is_available()\n If true then cuda is avilable"
                             + TextColor.END)
            exit(1)

        # Now see which devices to use
        if self.device_ids is None:
            total_gpu_devices = torch.cuda.device_count()
            sys.stderr.write(TextColor.GREEN + "INFO: TOTAL GPU AVAILABLE: " + str(total_gpu_devices) + "\n" + TextColor.END)
            device_ids = [i for i in range(0, total_gpu_devices)]
            callers = total_gpu_devices
        else:
            device_ids = [int(i) for i in self.device_ids.split(',')]
            for device_id in device_ids:
                major_capable, minor_capable = torch.cuda.get_device_capability(device=device_id)
                if major_capable < 0:
                    sys.stderr.write(TextColor.RED + "ERROR: GPU DEVICE: " + str(device_id) + " IS NOT CUDA CAPABLE.\n" + TextColor.END)
                    sys.stderr.write(TextColor.GREEN + "Try running: $ python3\n"
                                                       ">>> import torch \n"
                                                       ">>> torch.cuda.get_device_capability(device="
                                     + str(device_id) + ")\n" + TextColor.END)
                else:
                    sys.stderr.write(TextColor.GREEN + "INFO: CAPABILITY OF GPU#" + str(device_id)
                                     + ":\t" + str(major_capable) + "-" + str(minor_capable) + "\n" + TextColor.END)
            callers = len(device_ids)

        if callers == 0:
            sys.stderr.write(TextColor.RED + "ERROR: NO GPU AVAILABLE BUT GPU MODE IS SET\n" + TextColor.END)
            exit()

        # train a model
        train_distributed(self.train_file,
                          self.test_file,
                          self.batch_size,
                          self.epochs,
                          self.gpu_mode,
                          self.num_workers,
                          self.retrain_model,
                          self.retrain_model_path,
                          self.gru_layers,
                          self.hidden_size,
                          self.learning_rate,
                          self.weight_decay,
                          self.model_dir,
                          self.stats_dir,
                          device_ids,
                          callers,
                          train_mode=True)


def train_interface(train_dir, test_dir, gpu_mode, device_ids, epoch_size, batch_size, num_workers, output_dir,
                    retrain_model, retrain_model_path):
    """
    Interface to perform training
    :param train_dir: Path to directory containing training images
    :param test_dir: Path to directory containing training images
    :param gpu_mode: GPU mode
    :param device_ids: Device IDs of devices to use for GPU inference
    :param epoch_size: Number of epochs to train on
    :param batch_size: Batch size
    :param num_workers: Number of workers for data loading
    :param output_dir: Path to directory to save model
    :param retrain_model: If you want to retrain an existing model
    :param retrain_model_path: Path to the model you want to retrain
    :return:
    """
    model_out_dir, stats_dir = FileManager.handle_train_output_directory(output_dir)
    tm = TrainModule(train_dir,
                     test_dir,
                     gpu_mode,
                     device_ids,
                     epoch_size,
                     batch_size,
                     num_workers,
                     retrain_model,
                     retrain_model_path,
                     model_out_dir,
                     stats_dir)

    if gpu_mode:
        tm.train_model_gpu()
    else:
        tm.train_model()



import h5py
import yaml
import numpy as np


class DataStore(object):
    """
    This DataStore handles how we write an intermediate prediction file which we can use in stitch so we can
    do the stitching in parallel. The intended use of this object is to open under "with" statement.
    """
    # path to all the predictions in the HDF file.
    _prediction_path_ = 'predictions'
    # values that we save in the file
    _groups_ = ('position', 'index', 'bases', 'rles')

    def __init__(self, filename, mode='r'):
        """
        Object initialization function
        :param filename: Name of the function where to save the data
        :param mode: 'w' for write and 'r' for read.
        """
        # set the filename and mode
        self.filename = filename
        self.mode = mode

        self._sample_keys = set()
        # set the file handler
        self.file_handler = h5py.File(self.filename, self.mode)

        self._meta = None

    def __enter__(self):
        """
        This method is invoked when we open an object under "with" statement.
        :return:
        """
        self.file_handler = h5py.File(self.filename, self.mode)

        return self

    def __exit__(self, *args):
        """
        This method is invoked when the file object goes out of scope
        :param args:
        :return:
        """
        # we are not actively using the metadata now, we can remove this from the initial version
        if self.mode != 'r' and self._meta is not None:
            self.write_metadata(self.meta)
        self.file_handler.close()

    def write_metadata(self, data):
        """
        Save a data structure to file within a yml str.
        :param data: a dictionary containing metadata values
        """
        for group, d in data.items():
            if group in self.file_handler:
                del self.file_handler[group]
            self.file_handler[group] = yaml.dump(d)

    def load_metadata(self, groups=None):
        """
        Load meta data
        """
        if groups is None:
            groups = self._groups_
        return {g: yaml.load(self.file_handler[g][()]) for g in groups if g in self.file_handler}

    @property
    def meta(self):
        if self._meta is None:
            self._meta = self.load_metadata()
        return self._meta

    def update_meta(self, meta):
        """
        Update metadata
        """
        self._meta = self.meta
        self._meta.update(meta)

    def write_prediction(self, contig, contig_start, contig_end, chunk_id, position,
                         predicted_bases, predicted_rles, filename):
        """
        This is the method we use to write the data to the HDF file. This method is called by each image we
        generate.
        :param contig: Name of contig where the image belongs to.
        :param contig_start: Contig start position of the image.
        :param contig_end: Contig end position of the image.
        :param chunk_id: Chunk id from marginpolish to know which chunk the image is from.
        :param position: Array of values indicating genomic positions.
        :param predicted_bases: Array of values indicating predicted bases for each genomic position.
        :param predicted_rles: Array of values indicating predicted run-length for each genomic position.
        :param filename: Name of the file the image belongs to (used for debugging mostly)
        :return:
        """
        # get a chunk name prefix and suffix
        chunk_name_prefix = str(contig) + "-" + str(contig_start.item()) + "-" + str(contig_end.item())
        chunk_name_suffix = str(chunk_id.item())

        name = contig + chunk_name_prefix + chunk_name_suffix

        if 'predictions' not in self.meta:
            self.meta['predictions'] = set()
        if 'predictions_contig' not in self.meta:
            self.meta['predictions_contig'] = set()

        # the way we set up is predictions -> contig -> chunk_name_prefix -> chunk_name_suffix -> image
        # this way it's easier for stitch.py to grab all the images that belongs to a contig, split them by
        # chunk name prefix and then stitch the adjacent sequences together.

        # we set-up the contig -> chunk_name_prefix here. This sets up the contig start and end position for
        # marginpolish chunks
        if chunk_name_prefix not in self.meta['predictions_contig']:
            self.meta['predictions_contig'].add(chunk_name_prefix)
            self.file_handler['{}/{}/{}/{}'.format(self._prediction_path_, contig, chunk_name_prefix, 'contig_start')] \
                = contig_start.item()
            self.file_handler['{}/{}/{}/{}'.format(self._prediction_path_, contig, chunk_name_prefix, 'contig_end')] \
                = contig_end.item()

        # then we save the image predictions at the right place.
        if name not in self.meta['predictions']:
            self.meta['predictions'].add(name)
            self.file_handler['{}/{}/{}/{}/{}'.format(self._prediction_path_, contig, chunk_name_prefix,
                                                      chunk_name_suffix, 'position')] = np.array(position,
                                                                                                 dtype=np.uint32)
            self.file_handler['{}/{}/{}/{}/{}'.format(self._prediction_path_, contig, chunk_name_prefix,
                                                      chunk_name_suffix, 'bases')] = np.array(predicted_bases,
                                                                                              dtype=np.uint8)
            self.file_handler['{}/{}/{}/{}/{}'.format(self._prediction_path_, contig, chunk_name_prefix,
                                                      chunk_name_suffix, 'rles')] = np.array(predicted_rles,
                                                                                             dtype=np.uint8)



import h5py
import argparse
from collections import defaultdict, OrderedDict
import yaml
from tqdm import tqdm

helen_positional_features = defaultdict(list)
helen_positional_labels = defaultdict(int)
all_helen_positions = set()
all_medaka_positions = set()
_gap_ = '*'
_ref_gap_ = '#'
_read_sep_ = ' '
_alphabet_ = 'ACGT'
_extra_bases_ = 'NRYSWKMBDHV'
decoding = _gap_ + _alphabet_.lower() + _alphabet_.upper() + _read_sep_ + _extra_bases_
encoding = OrderedDict(((a, i) for i, a in enumerate(decoding)))
helen_decoding = {0: '*', 1: 'A', 2: 'C', 3: 'G', 4: 'T', 5: '*'}


def read_helen_h5py(hdf5_image, with_labels):
    hdf5_file = h5py.File(hdf5_image, 'r')

    image_dataset = hdf5_file['image']
    position_dataset = hdf5_file['position']
    index_dataset = hdf5_file['index']

    label_dataset = None
    if with_labels:
        label_dataset = hdf5_file['label']

    total_records = image_dataset.shape[0]
    print("PARSING HELEN DICTIONARY")
    for hdf5_index in tqdm(range(0, total_records), ncols=100):

        position = position_dataset[hdf5_index]
        index = index_dataset[hdf5_index]
        image = image_dataset[hdf5_index]

        if with_labels:
            label = label_dataset[hdf5_index]
            for l, p, i, f in zip(label, position, index, image):
                helen_positional_features[(p, i)] = f
                helen_positional_labels[(p, i)] = l
                all_helen_positions.add((p, i))
        else:
            for p, i, f in zip(position, index, image):
                if p < 0 or i < 0:
                    continue
                helen_positional_features[(p, i)] = f
                all_helen_positions.add((p, i))


def comapare_medaka_h5py(medaka_h5py, with_labels):
    hdf5_file = h5py.File(medaka_h5py, 'r')
    used_labels_list = list(yaml.load(hdf5_file['medaka_feature_decoding'][()]))

    if with_labels:
        used_labels_list = [(base, count) for used, base, count in used_labels_list if used == True]
        map_labels_list = list(yaml.load(hdf5_file['medaka_label_counts'][()]))

    samples = hdf5_file['samples']
    with tqdm(total=len(samples), leave=True, ncols=100) as progress_bar:
        for sample in samples:
            total_records = hdf5_file.get('/samples/'+sample+'/features').shape[0]

            features_dataset = hdf5_file.get('/samples/'+sample+'/features')
            positions_dataset = hdf5_file.get('/samples/'+sample+'/positions')

            labels_dataset = None

            if with_labels:
                labels_dataset = hdf5_file.get('/samples/'+sample+'/labels')

            for i in range(0, total_records):
                medaka_features = list(features_dataset[i])
                position = tuple(positions_dataset[i])

                # compare features
                if position not in helen_positional_features.keys():
                    # pass
                    print("POSITION NOT PRESENT IN HELEN: ", position)
                else:
                    # remove from helen's set
                    all_medaka_positions.add(position)
                    # compare encoded feature's values
                    helen_features = helen_positional_features[position]
                    equal = True
                    max_diff = max(abs(helen_features - medaka_features))
                    if max_diff > 0.0001:
                        equal = False
                    if not equal:
                        print("FEATURES DID NOT MATCH IN POSITION: ", position)
                        print("MEDAKA FEATURES: ", ',\t'.join(str(round(x, 3)) for x in medaka_features))
                        print("HELEN  FEATURES: ", ',\t'.join(str(round(x, 3)) for x in helen_features))

                    # compare the labels
                    if with_labels:
                        label, count = tuple(labels_dataset[i])
                        label_base = decoding[label]
                        medaka_label = label_base
                        helen_label = helen_decoding[helen_positional_labels[position]]
                        if helen_label != medaka_label:
                            print("LABELS DID NOT MATCH IN POSITION: ", position, "HELEN LABEL: ", helen_label,
                                  "MEDAKA LABEL", medaka_label)
            progress_bar.refresh()
            progress_bar.update(1)
        progress_bar.close()

    print("HELEN EXTRA POSITIONS: ", all_helen_positions.difference(all_medaka_positions))
    print("MEDAKA EXTRA POSITIONS: ", all_medaka_positions.difference(all_helen_positions))


if __name__ == '__main__':
    '''
    Processes arguments and performs tasks.
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--helen_h5py",
        type=str,
        required=True,
        help="H5PY file generated by HELEN."
    )
    parser.add_argument(
        "--medaka_h5py",
        type=str,
        required=True,
        help="H5PY file generated by MEDAKA."
    )
    parser.add_argument(
        "--with_label",
        type=lambda x: (str(x).lower() == 'true' or str(x).lower() == '1'),
        default=False,
        help="If true then compare labels too."
    )

    FLAGS, unparsed = parser.parse_known_args()
    read_helen_h5py(FLAGS.helen_h5py, FLAGS.with_label)
    comapare_medaka_h5py(FLAGS.medaka_h5py, FLAGS.with_label)



import argparse
import pysam
from collections import defaultdict


def fix_vcf(input_vcf, output_vcf):
    vcf_in = pysam.VariantFile(input_vcf)
    vcf_out = pysam.VariantFile(output_vcf, 'w', header=vcf_in.header)
    records = vcf_in.fetch()
    PS_dictionary = defaultdict(str)
    PS_value = 100

    for record in records:
        for sample in record.samples:
            input_ps = str(record.samples[sample]['PS'])

            if input_ps not in PS_dictionary:
                PS_dictionary = str(PS_value)
                PS_value += 50

            record.samples[sample]['PS'] = str(PS_value)
            vcf_out.write(record)


if __name__ == '__main__':
    '''
    Processes arguments and performs tasks.
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input_vcf",
        type=str,
        required=True,
        help="H5PY file generated by HELEN."
    )
    parser.add_argument(
        "--output_vcf",
        type=str,
        required=True,
        help="H5PY file generated by HELEN."
    )

    FLAGS, unparsed = parser.parse_known_args()
    fix_vcf(FLAGS.input_vcf, FLAGS.output_vcf)



import math
import argparse
import torch
import numpy as np
import h5py
from torchvision import transforms


def get_strand_color(is_rev):
    """
    Get color for forward and reverse reads
    :param is_rev: True if read is reversed
    :return:
    """
    is_rev = int(math.ceil(is_rev))
    if is_rev == 254:
        return 'R'
    if is_rev == 240:
        return '1'
    elif is_rev == 70:
        return '0'
    else:
        return ' '


def get_alt_type(alt_type_color):
    """
    Get color for forward and reverse reads
    :param is_rev: True if read is reversed
    :return:
    """
    alt_type_color = int(math.ceil(alt_type_color))
    if alt_type_color == 0:
        return ' '
    elif alt_type_color == 5:
        return '0'
    elif alt_type_color == 240:
        return '1'
    elif alt_type_color == 125:
        return '2'
    elif alt_type_color == 254:
        return 'R'


def get_base_from_color(base_color):
    color = int(math.ceil(base_color))
    global_base_color_reverse = {200: 'A', 50: 'C', 150: 'G', 100: 'T', 10: '.', 250: '*'}
    # {{'C', 50}, {'T', 100}, {'G', 150}, {'A', 200}, {'*', 250}, {'.', 10}, {'N', 10}};
    if color in global_base_color_reverse:
        return global_base_color_reverse[color]
    else:
        return ' '
    # 'A': 25.0, 'C': 75.0, 'G': 125.0, 'T': 175.0, '*': 225.0


def get_quality_by_color(quality):
    """
    Get a color spectrum given mapping quality
    :param map_quality: value of mapping quality
    :return:
    """
    quality = int(math.ceil(quality))
    color = math.floor(((quality / 254) * 9))
    if color == 0:
        return ' '
    return str(color)


def get_mismatch_or_alt_color(alt_type_color):
    """
    Get color for forward and reverse reads
    :param is_rev: True if read is reversed
    :return:
    """
    alt_type_color = int(math.ceil(alt_type_color))
    if alt_type_color == 0:
        return ' '
    elif alt_type_color == 50:
        return '0'
    elif alt_type_color == 250:
        return '1'


def analyze_tensor(image):
    # base_color, base_quality_color, map_qual_color, strand_color, alt_color
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    print("BASE CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_base_from_color(image[0][j][i]), end='')
        print()

    print("BASE QUALITY CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_quality_by_color(image[1][j][i]), end='')
        print()
    print("MAPPING QUALITY CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_quality_by_color(image[2][j][i]), end='')
        print()
    print("STRAND DIRECTION CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_strand_color(image[3][j][i]), end='')
        print()
    print("MISMATCH CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_mismatch_or_alt_color(image[4][j][i]), end='')
        print()
    print("ALT1 CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_mismatch_or_alt_color(image[5][j][i]), end='')
        print()
    print("ALT2 CHANNEL:")
    for i in range(img_h):
        for j in range(img_w):
            print(get_mismatch_or_alt_color(image[6][j][i]), end='')
        print()


def save_base_quality_array(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            image_row.append([0, 0, image[1][j][i], 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("base_quality_tensor" + ".png", entire_image, format="PNG")


def save_map_quality_array(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            image_row.append([0, image[2][j][i], 0, 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("map_quality_tensor" + ".png", entire_image, format="PNG")


def save_strand_array(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            image_row.append([image[3][j][i], 0, 0, 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("strand_color_tensor" + ".png", entire_image, format="PNG")


def save_mismatch(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            image_row.append([0, image[4][j][i], 0, 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("mismatch_tensor" + ".png", entire_image, format="PNG")


def save_alt1(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            image_row.append([image[5][j][i], 0, 0, 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("alt1_freq_tensor" + ".png", entire_image, format="PNG")


def save_alt2(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            image_row.append([0, 0, image[6][j][i], 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("alt2_freq_tensor" + ".png", entire_image, format="PNG")


def save_base_array(image):
    img_c, img_w, img_h = image.size()
    image = np.array(image.data * 254)
    img_h = 100
    entire_image = []
    for i in range(img_h):
        image_row = []
        for j in range(img_w):
            if image[0][j][i] != 0:
                print(get_base_from_color(image[0][j][i]), end='')
                if get_base_from_color(image[0][j][i]) == ' ':
                    image_row.append([255, 255, 255, 255])
                # elif get_base_from_color(image[0][j][i]) == get_base_from_color(image[0][j][0]) and i > 0:
                #     image_row.append([255, 255, 255, 255])
                elif get_base_from_color(image[0][j][i]) == 'A':
                    image_row.append([0, 0, 255, 255])
                elif get_base_from_color(image[0][j][i]) == 'C':
                    image_row.append([255, 0, 0, 255])
                elif get_base_from_color(image[0][j][i]) == 'G':
                    image_row.append([0, 255, 0, 255])
                elif get_base_from_color(image[0][j][i]) == 'T':
                    image_row.append([255, 255, 0, 255])
                else:
                    # purple
                    image_row.append([160, 32, 240, 255])
            else:
                print(' ', end='')
                image_row.append([0, 0, 0, 255])
        entire_image.append(image_row)
        print()
    entire_image = np.array(entire_image)
    from scipy import misc
    misc.imsave("base_tensor" + ".png", entire_image, format="PNG")


def tensor_to_image(image):
    # base_color, base_quality_color, map_qual_color, strand_color, alt_color
    print("BASE CHANNEL:")
    save_base_array(image)
    save_base_quality_array(image)
    save_map_quality_array(image)
    save_strand_array(image)
    save_mismatch(image)
    save_alt1(image)
    save_alt2(image)


if __name__ == '__main__':
    '''
    Processes arguments and performs tasks.
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tensor_file",
        type=str,
        required=True,
        help="H5PY file path"
    )
    parser.add_argument(
        "--index",
        type=int,
        required=True,
        help="Index of image."
    )
    FLAGS, unparsed = parser.parse_known_args()

    if FLAGS.tensor_file:
        hdf5_image = FLAGS.tensor_file
        hdf5_index = FLAGS.index
        hdf5_file = h5py.File(hdf5_image, 'r')

        image_dataset = hdf5_file['images']
        image = np.array(image_dataset[hdf5_index], dtype=np.uint8)
        transform = transforms.Compose([transforms.ToTensor()])
        image = transform(image)
        image = image.transpose(1, 2)

        label_dataset = hdf5_file['labels']
        label = np.array(label_dataset[hdf5_index], dtype=np.long)
        for l in label:
            print(l, end='')
        print()
        # analyze_tensor(image)

        tensor_to_image(image)




import wget
import sys
from modules.python.FileManager import FileManager
from modules.python.TextColor import TextColor


def download_models(output_dir):
    output_dir = FileManager.handle_output_directory(output_dir)
    sys.stderr.write(TextColor.YELLOW + "DOWNLOADING MODEL DESCRIPTION FILE" + TextColor.END + "\n")
    description_file = "https://storage.googleapis.com/kishwar-helen/models_helen/model_description.csv"
    wget.download(description_file, output_dir)
    sys.stderr.write("\n")
    sys.stderr.flush()

    with open(output_dir+'/model_description.csv') as f:
        models = [line.rstrip() for line in f]

    for model in models:
        model_name, model_url = model.split(',')
        sys.stderr.write("INFO: DOWNLOADING FILE: " + str(model_name) + ".pkl\n")
        sys.stderr.write("INFO: DOWNLOADING LINK: " + str(model_url) + "\n")
        wget.download(model_url, output_dir)
        sys.stderr.write("\n")
        sys.stderr.flush()


import os
import time


class FileManager:
    """
    Performs simple OS operations like creating directory for output or path to all the files.
    """
    @staticmethod
    def handle_output_directory(output_dir):
        """
        Process the output directory and return a valid directory where we save the output
        :param output_dir: Output directory path
        :return:
        """
        # process the output directory
        if output_dir[-1] != "/":
            output_dir += "/"
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        return os.path.abspath(output_dir)

    @staticmethod
    def handle_train_output_directory(output_dir):
        """
        Process the output directory and return a valid directory where we save the output
        :param output_dir: Output directory path
        :return:
        """
        timestr = time.strftime("%m%d%Y_%H%M%S")
        # process the output directory
        if output_dir[-1] != "/":
            output_dir += "/"
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        # create an internal directory so we don't overwrite previous runs
        model_save_dir = output_dir + "trained_models_" + timestr + "/"
        if not os.path.exists(model_save_dir):
            os.mkdir(model_save_dir)

        stats_directory = model_save_dir + "stats_" + timestr + "/"

        if not os.path.exists(stats_directory):
            os.mkdir(stats_directory)

        return model_save_dir, stats_directory

    @staticmethod
    def get_file_paths_from_directory(directory_path):
        """
        Returns all paths of .h5 files given a directory path
        :param directory_path: Path to the directory
        :return: A list of paths of files
        """
        file_paths = [os.path.join(directory_path, file) for file in os.listdir(directory_path)
                      if os.path.isfile(os.path.join(directory_path, file)) and file[-2:] == 'h5']
        return file_paths

    @staticmethod
    def chunks(file_names, threads):
        """
        Given a list of file names a number of threads, this method returns a list containing file names.
        The len(chunks) == threads so we can use each item in the list for a single process.
        """
        chunks = []
        for i in range(0, len(file_names), threads):
            chunks.append(file_names[i:i + threads])
        return chunks




import argparse
import sys
import torch
from version import __version__
from modules.python.TextColor import TextColor
from modules.python.TrainInterface import train_interface
from modules.python.TestInterface import test_interface


def add_train_arguments(parser):
    """
    Add arguments to a parser for sub-command "train"
    :param parser: argeparse object
    :return:
    """
    parser.add_argument(
        "--train_image_dir",
        type=str,
        required=True,
        help="Path to directory containing labeled images for training."
    )
    parser.add_argument(
        "--test_image_dir",
        type=str,
        required=True,
        help="Path to directory containing labeled images for testing the models."
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        required=False,
        default=100,
        help="Batch size for training, default is 100."
    )
    parser.add_argument(
        "--epoch_size",
        type=int,
        required=False,
        default=10,
        help="Epoch size for training iteration."
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        required=False,
        default='./model',
        help="Output directory path for model to save."
    )
    parser.add_argument(
        "--retrain_model",
        type=bool,
        default=False,
        help="If true then retrain a pre-trained mode."
    )
    parser.add_argument(
        "--retrain_model_path",
        type=str,
        default=False,
        help="Path to the model that will be retrained."
    )
    parser.add_argument(
        "--gpu_mode",
        default=False,
        action='store_true',
        help="If set then PyTorch will use GPUs for inference. CUDA required."
    )
    parser.add_argument(
        "-d_ids",
        "--device_ids",
        type=str,
        required=False,
        default=None,
        help="List of gpu device ids to use for inference.\n"
             "Example usage: --device_ids 0,1,2 (this will create three callers in id 'cuda:0, cuda:1 and cuda:2'\n"
             "If none then it will use all available devices."
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        required=False,
        default=16,
        help="Number of workers to assign to the dataloader."
    )
    return parser


def add_test_arguments(parser):
    """
    Add arguments to a parser for sub-command "test"
    :param parser: argeparse object
    :return:
    """
    parser.add_argument(
        "--test_image_dir",
        type=str,
        required=True,
        help="Path to directory containing labeled images for testing the models."
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        required=False,
        default=100,
        help="Batch size for training, default is 100."
    )
    parser.add_argument(
        "--model_path",
        type=str,
        required=False,
        default='./model',
        help="Path of the model to load and retrain"
    )
    parser.add_argument(
        "--gpu_mode",
        action='store_true',
        help="If true then cuda is on."
    )
    parser.add_argument(
        "--print_details",
        action='store_true',
        help="If set then print details on mismatch cases."
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        required=False,
        default='./debug_output',
        help="Output file name."
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        required=False,
        default=40,
        help="Epoch size for training iteration."
    )
    return parser


def add_hyperband_arguments(parser):
    """
    Add arguments to a parser for sub-command "stitch"
    :param parser: argeparse object
    :return:
    """
    parser.add_argument(
        "--train_image_dir",
        type=str,
        required=True,
        help="Path to directory containing labeled images for training."
    )
    parser.add_argument(
        "--test_image_dir",
        type=str,
        required=True,
        help="Path to directory containing labeled images for testing the models."
    )
    parser.add_argument(
        "--gpu_mode",
        action='store_true',
        help="If true then cuda is on."
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        required=False,
        default='./debug_output',
        help="Output file name."
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        required=False,
        default=100,
        help="Batch size for training, default is 100."
    )
    parser.add_argument(
        "--max_epochs",
        type=int,
        required=False,
        default=10,
        help="Epoch size for training iteration."
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        required=False,
        default=40,
        help="Epoch size for training iteration."
    )
    return parser


def main():
    """
    Main interface for HELEN training models:
    1) Train
    2) Test
    3) Hyperband
    """
    parser = argparse.ArgumentParser(
        description="The train module of HELEN trains a deep neural network to perform a multi-task classification. "
                    "It takes a set of labeled images from MarginPolish and trains the model to predict a base and "
                    "the run-length of the base using a gated recurrent unit (GRU) based model. "
                    "This script is the interface to the training module.",
                    formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        "--version",
        default=False,
        action='store_true',
        help="Show version."
    )

    subparsers = parser.add_subparsers(dest='sub_command')
    # subparsers.required = True

    parser_train = subparsers.add_parser('train', help="Train a HELEN model. Requires a set of labeled images.")
    add_train_arguments(parser_train)

    parser_test = subparsers.add_parser('test', help="Test a model. Requires a set of labeled images")
    add_test_arguments(parser_test)

    subparsers.add_parser('torch_stat', help="See PyTorch configuration.")
    subparsers.add_parser('version', help="Show program version.")

    FLAGS, unparsed = parser.parse_known_args()

    if FLAGS.sub_command == 'train':
        sys.stderr.write(TextColor.GREEN + "INFO: TRAIN MODULE SELECTED\n" + TextColor.END)
        train_interface(FLAGS.train_image_dir,
                        FLAGS.test_image_dir,
                        FLAGS.gpu_mode,
                        FLAGS.device_ids,
                        FLAGS.epoch_size,
                        FLAGS.batch_size,
                        FLAGS.num_workers,
                        FLAGS.output_dir,
                        FLAGS.retrain_model,
                        FLAGS.retrain_model_path)

    elif FLAGS.sub_command == 'test':
        sys.stderr.write(TextColor.GREEN + "INFO: TEST MODULE SELECTED\n" + TextColor.END)
        test_interface(FLAGS.test_image_dir,
                       FLAGS.batch_size,
                       FLAGS.gpu_mode,
                       FLAGS.num_workers,
                       FLAGS.model_path,
                       FLAGS.output_dir,
                       FLAGS.print_details)

    elif FLAGS.sub_command == 'torch_stat':
        sys.stderr.write(TextColor.YELLOW + "TORCH VERSION: " + TextColor.END + str(torch.__version__) + "\n\n")
        sys.stderr.write(TextColor.YELLOW + "PARALLEL CONFIG:\n" + TextColor.END)
        print(torch.__config__.parallel_info())
        sys.stderr.write(TextColor.YELLOW + "BUILD CONFIG:\n" + TextColor.END)
        print(*torch.__config__.show().split("\n"), sep="\n")

        sys.stderr.write(TextColor.GREEN + "CUDA AVAILABLE: " + TextColor.END + str(torch.cuda.is_available()) + "\n")
        sys.stderr.write(TextColor.GREEN + "GPU DEVICES: " + TextColor.END + str(torch.cuda.device_count()) + "\n")

    elif FLAGS.version is True:
        print("HELEN VERSION: ", __version__)

    else:
        sys.stderr.write(TextColor.RED + "ERROR: NO SUBCOMMAND SELECTED. "
                                         "PLEASE SELECT ONE OF THE AVAIABLE SUB-COMMANDS.\n"
                         + TextColor.END)
        parser.print_help()


if __name__ == '__main__':
    main()


from torch.utils.data import Dataset
import torchvision.transforms as transforms
import h5py
import sys
from modules.python.TextColor import TextColor
from modules.python.FileManager import FileManager


class SequenceDataset(Dataset):
    """
    This class implements the dataset class for the dataloader to use.
    This version is intended to use with train.py and test.py as this method loads labels with the images.
    It initializes all the given images and returns each image through __getitem__.
    """
    def __init__(self, image_directory):
        """
        This method initializes the dataset by loading all the image information. It creates a sequential list
        call all_images from which we can grab images iteratively through __getitem__.
        :param image_directory: Path to a directory where all the images are saved.
        """
        # transformer to convert loaded objects to tensors
        self.transform = transforms.Compose([transforms.ToTensor()])

        # a list of file-image pairs, where we have (file_name, image_name) as values so we can fetch images
        # from the list of files.
        file_image_pair = []

        # get all the h5 files that we have in the directory
        hdf_files = FileManager.get_file_paths_from_directory(image_directory)
        for hdf5_file_path in hdf_files:
            # for each of the files get all the images
            with h5py.File(hdf5_file_path, 'r') as hdf5_file:
                # check if marginpolish somehow generated an empty file
                if 'images' in hdf5_file:
                    image_names = list(hdf5_file['images'].keys())

                    # save the file-image pair to the list
                    for image_name in image_names:
                        file_image_pair.append((hdf5_file_path, image_name))
                else:
                    sys.stderr.write(TextColor.YELLOW + "WARN: NO IMAGES FOUND IN FILE: "
                                     + hdf5_file_path + "\n" + TextColor.END)

        # save the list to all_images so we can access the list inside other methods
        self.all_images = file_image_pair

    def __getitem__(self, index):
        """
        This method returns a single object. Dataloader uses this method to load images and then minibatches the loaded
        images
        :param index: Index indicating which image from all_images to be loaded
        :return:
        """
        # get the file path and the name of the image
        hdf5_filepath, image_name = self.all_images[index]

        # load the image and the label
        with h5py.File(hdf5_filepath, 'r') as hdf5_file:
            image = hdf5_file['images'][image_name]['image'][()]
            label_base = hdf5_file['images'][image_name]['label_base'][()]
            label_run_length = hdf5_file['images'][image_name]['label_run_length'][()]

        return image, label_base, label_run_length

    def __len__(self):
        """
        Returns the length of the dataset
        :return:
        """
        return len(self.all_images)



from build import HELEN
from modules.python.TextColor import TextColor
from modules.python.Options import StitchOptions
from collections import defaultdict


class PairWiseAlignment:

    def perform_pairwise_alignment(self, ref, query):

        aligner = HELEN.Aligner(StitchOptions.MATCH_PENALTY, StitchOptions.MISMATCH_PENALTY,
                                StitchOptions.GAP_PENALTY, StitchOptions.GAP_EXTEND_PENALTY)

        # first create a k-mer index for the ref, like minimap
        ref_kmers = defaultdict(list)
        for i in range(len(ref)):
            kmer = ref[i:i+StitchOptions.KMER_SIZE]

            if len(kmer) < StitchOptions.KMER_SIZE:
                break

            ref_kmers[kmer].append(i)

        previous_match_positions = []
        for query_index in range(0, len(query), 5):
            query_kmer = query[query_index:query_index+StitchOptions.KMER_SIZE]

            if len(query_kmer) < StitchOptions.KMER_SIZE:
                break
            ref_index_matches = ref_kmers[query_kmer]

            for ref_index in ref_index_matches:
                if ref_index - 5 not in previous_match_positions:
                    print("UNIQUE", query_index, ref_index)
                    sub_query_seq = query[query_index:]
                    sub_ref_seq = ref[ref_index:]

                    alignment = HELEN.Alignment()
                    aligner.SetReferenceSequence(sub_ref_seq, len(sub_ref_seq))
                    filter_profile = HELEN.Filter()
                    aligner.Align_cpp(sub_query_seq, filter_profile, alignment, 0)

                    print(sub_ref_seq)
                    print(sub_query_seq)
                    print(alignment.best_score)
                    print(alignment.cigar_string)
                    print("-----------------")

            previous_match_positions = ref_index_matches

        exit()




        # initialize an alignment object

        # align current sequence to the previous sequence
        filter_profile = HELEN.Filter()
        aligner.Align_cpp(query, filter_profile, alignment, 0)

        print(alignment.cigar_string)
        print(alignment.best_score)
        print(alignment.reference_begin)
        print(alignment.query_begin)





from modules.python.PairwiseAlignment import PairWiseAlignment

a = PairWiseAlignment()
ref = "TAGTATATACACTATACTATATACTAGTGTATTATACTAGTATATACACTATACTATATACTAGTGTATTATACTAGTATATACACTATACTATATACTAGTGTATATACTAGTATATAC"
query = "TATACTAGTGTATTATACTAGTATAATACACTATACTATATACTAGTGTATTATACTAGTATATACACTATACTATATGCTAGTGTATTATACTAGTATATACACTATACTATATACTAG"

a.perform_pairwise_alignment(ref, query)


class StitchOptions(object):
    BASE_ERROR_RATE = 0.0
    label_decoder = {1: 'A', 2: 'C', 3: 'G', 4: 'T', 0: ''}
    MATCH_PENALTY = 4
    MISMATCH_PENALTY = 6
    GAP_PENALTY = 8
    GAP_EXTEND_PENALTY = 2
    MIN_SEQUENCE_REQUIRED_FOR_MULTITHREADING = 2
    OVERLAP_THRESHOLD = 8
    KMER_SIZE = 15


class ImageSizeOptions(object):
    IMAGE_HEIGHT = 90
    IMAGE_CHANNELS = 1
    SEQ_LENGTH = 1000
    SEQ_OVERLAP = 200
    LABEL_LENGTH = SEQ_LENGTH

    TOTAL_BASE_LABELS = 5
    TOTAL_RLE_LABELS = 11


class TrainOptions(object):
    TRAIN_WINDOW = 100
    WINDOW_JUMP = 50
    GRU_LAYERS = 1
    HIDDEN_SIZE = 128
    CLASS_WEIGHTS = [0.3, 0.5, 0.5, 0.5, 0.5, 0.8, 0.9, 1.0, 1.0, 1.0, 0.9]




import subprocess as sp
import sys
import os
"""
This is a wrapper to call MarginPolish from the installed binary directory.
"""


def main():
    """
    Run MarginPolish from inside the binary directory
    :return:
    """
    command = os.path.join(os.path.dirname(__file__) + "/bin/marginPolish")
    sp.call([command] + sys.argv)


if __name__ == '__main__':
    main()



import sys
import time
from modules.python.TextColor import TextColor
from modules.python.CallConsensusInterface import call_consensus
from modules.python.StitchInterface import perform_stitch
from modules.python.FileManager import FileManager
"""
The Call Consensus method generates base predictions for images generated through MarginPolish. This script reads
hdf5 files generated by MarginPolish and produces another Hdf5 file that holds all predictions. The generated hdf5 file
is given to stitch.py which then stitches the segments using an alignment which gives us a polished sequence.

The algorithm is described here:

  1) INPUTS:
    - directory path to the image files generated by MarginPolish
    - model path directing to a trained model
    - batch size for mini-batch prediction
    - num workers for mini-batch processing threads
    - output directory path to where the output hdf5 will be saved
    - gpu mode indicating if GPU will be used
  2) METHOD:
    - Call predict function that loads the neural network and generates base predictions and saves it into a hdf5 file
        - Loads the model
        - Iterates over the input images in minibatch
        - For each image uses a sliding window method to slide of the image sequence
        - Aggregate the predictions to get sequence prediction for the entire image sequence
        - Save all the predictions to a file
  3) OUTPUT:
    - A hdf5 file containing all the base predictions   
"""


def get_elapsed_time_string(start_time, end_time):
    """
    Get a string representing the elapsed time given a start and end time.
    :param start_time: Start time (time.time())
    :param end_time: End time (time.time())
    :return:
    """
    elapsed = end_time - start_time
    hours = str(int(elapsed/3600))
    elapsed = elapsed - (elapsed/3600)
    mins = str(int(elapsed/60))
    secs = str(int(elapsed) % 60)
    time_string = hours + " HOURS " + mins + " MINS " + secs + " SECS."

    return time_string


def polish_genome(image_dir, model_path, batch_size, num_workers, threads, output_dir, output_prefix, gpu_mode,
                  device_ids, callers):
    """
    This method provides an interface too call the predict method that generates the prediction hdf5 file.
    :param image_dir: Path to directory where all MarginPolish images are saved.
    :param model_path: Path to a trained model.
    :param batch_size: Batch size for minibatch processing.
    :param num_workers: Number of workers for minibatch processing.
    :param threads: Number of threads for pytorch.
    :param output_dir: Path to the output directory.
    :param output_prefix: Prefix of the output HDF5 file.
    :param gpu_mode: If true, predict method will use GPU.
    :param device_ids: List of GPU devices.
    :param callers: Total number of callers to use.
    :return:
    """
    output_dir = FileManager.handle_output_directory(output_dir)
    timestr = time.strftime("%m%d%Y_%H%M%S")

    prediction_output_directory = output_dir + "/predictions_" + str(timestr) + "/"
    prediction_output_directory = FileManager.handle_output_directory(prediction_output_directory)

    sys.stderr.write(TextColor.GREEN + "INFO: RUN-ID: " + str(timestr) + "\n" + TextColor.END)
    sys.stderr.write(TextColor.GREEN + "INFO: PREDICTION OUTPUT DIRECTORY: "
                     + str(prediction_output_directory) + "\n" + TextColor.END)

    call_consensus_start_time = time.time()
    sys.stderr.write(TextColor.GREEN + "INFO: CALL CONSENSUS STARTING\n" + TextColor.END)
    call_consensus(image_dir,
                   model_path,
                   batch_size,
                   num_workers,
                   threads,
                   prediction_output_directory,
                   output_prefix,
                   gpu_mode,
                   device_ids,
                   callers)
    call_consensus_end_time = time.time()

    stitch_start_time = time.time()
    sys.stderr.write(TextColor.GREEN + "INFO: STITCH STARTING\n" + TextColor.END)
    print(prediction_output_directory)
    perform_stitch(prediction_output_directory,
                   output_dir,
                   output_prefix,
                   threads)
    stitch_end_time = time.time()

    call_consensus_time = get_elapsed_time_string(call_consensus_start_time, call_consensus_end_time)
    stitch_time = get_elapsed_time_string(stitch_start_time, stitch_end_time)
    overall_time = get_elapsed_time_string(call_consensus_start_time, stitch_end_time)

    sys.stderr.write(TextColor.GREEN + "INFO: FINISHED PROCESSING.\n" + TextColor.END)
    sys.stderr.write(TextColor.GREEN + "INFO: TOTAL TIME ELAPSED: " + str(overall_time) + "\n" + TextColor.END)
    sys.stderr.write(TextColor.GREEN + "INFO: PREDICTION TIME: " + str(call_consensus_time) + "\n" + TextColor.END)
    sys.stderr.write(TextColor.GREEN + "INFO: STITCH TIME: " + str(stitch_time) + "\n" + TextColor.END)




import torch
import torch.nn as nn
import warnings

# this ignore is to avoid the flatten parameter warning from showing up. There is no way around it on GPU but it
# still shows up.
warnings.filterwarnings("ignore", category=RuntimeWarning)
"""
This script defines a pytorch deep neural network model for a multi-task classification problem that we 
use in HELEN.

MODEL DESCRIPTION:
The model is a simple bidirectional gated recurrent unit (GRU) based encoder-decoder model. The first layer
performs encoding of the input features and the second layer does a decoder. We use two separate linear layer
to perform the base prediction and RLE prediction. This multi-task method uses a parameter sharing scheme which
is very popular in multi-task classification problems.
"""


class TransducerGRU(nn.Module):
    """
    The GRU based transducer model class.
    """
    def __init__(self, image_channels, image_features, gru_layers, hidden_size, num_base_classes, num_rle_classes,
                 bidirectional=True):
        """
        The initialization of the model
        :param image_channels: Number of channels, usually 1 so not used in this case.
        :param image_features: Features per column in the image
        :param gru_layers: Number of GRU layers in the model
        :param hidden_size: Size of the hidden layer
        :param num_base_classes: Number of classes in base prediction
        :param num_rle_classes: Number of classes in RLE prediction
        :param bidirectional: If true then the GRU layers will be bidirectional.
        """
        super(TransducerGRU, self).__init__()
        self.hidden_size = hidden_size
        self.bidirectional = bidirectional
        self.num_layers = gru_layers
        self.num_base_classes = num_base_classes
        self.num_rle_classes = num_rle_classes
        # the gru encoder layer
        self.gru_encoder = nn.GRU(image_features,
                                  hidden_size,
                                  num_layers=self.num_layers,
                                  bidirectional=bidirectional,
                                  batch_first=True)
        # the gru decoder layer
        self.gru_decoder = nn.GRU(2 * hidden_size,
                                  hidden_size,
                                  num_layers=self.num_layers,
                                  bidirectional=bidirectional,
                                  batch_first=True)
        self.gru_encoder.flatten_parameters()
        self.gru_decoder.flatten_parameters()
        # the linear layer for base and RLE classification
        self.dense1_base = nn.Linear(self.hidden_size * 2, self.num_base_classes)
        self.dense2_rle = nn.Linear(self.hidden_size * 2, self.num_rle_classes)

    def forward(self, x, hidden):
        """
        The forward method of the model.
        :param x: Input image
        :param hidden: Hidden input
        :return:
        """
        # this needs to ensure consistency between GPU/CPU training
        hidden = hidden.transpose(0, 1).contiguous()
        # encoding
        x_out_layer1, hidden_out_layer1 = self.gru_encoder(x, hidden)
        # decoding
        x_out_final, hidden_final = self.gru_decoder(x_out_layer1, hidden_out_layer1)

        # classification
        base_out = self.dense1_base(x_out_final)
        rle_out = self.dense2_rle(x_out_final)

        hidden_final = hidden_final.transpose(0, 1).contiguous()
        return base_out, rle_out, hidden_final

    def init_hidden(self, batch_size, num_layers, bidirectional=True):
        """
        Initialize the hidden tensor.
        :param batch_size: Size of the batch
        :param num_layers: Number of GRU layers
        :param bidirectional: If true then GRU layers are bidirectional
        :return:
        """
        num_directions = 1
        if bidirectional:
            num_directions = 2

        return torch.zeros(batch_size, num_directions * num_layers, self.hidden_size)
